# -*- coding: utf-8 -*-
"""
Created on Fri Aug  5 14:47:38 2016

@author: akibdpt
"""

from .form import form

from copy import deepcopy

"""
 * A multiply-form denotes a form of a multiply-associated sort.
 * Multiply-forms may contain any number of individuals.
 * Different types of multiply-forms are distinguished based on the operational
 * behavior of their forms. 
 *
 * A {DiscreteForm} specifies a discrete behavior corresponding to a set: any 
 * two individuals are either identical or disjoint.
 * An {IntervalForm} specifies a behavior of intervals on a
 * one-dimensional axis: if two intervals overlap, these are combined into a
 * single interval.
 * An {ArealForm} applies to two-dimensional planar segments and
 * three-dimensional volumes, and uses a boundary classification approach to
 * determine overlapping segment parts.
 * A {RelationalForm} applies to (bi-directional) relations. While its
 * behavior is that of a discrete form, relations are ultimately dependent on
 * the individuals these relate and are removed if any of their associate
 * individuals are removed.
 *
 * A multiply-form is maximal if no two individuals combine into a single
 * individual under the respective operational behavior, and if no individuals
 * equal to nil.
 * 
 * The MultiplyForm class extends on the {Form} class.
 * It implements the form container as an ordered {TreeSet} and maintains
 * for each form a reference to the attribute sort to which it belongs.
 * It provides common implementations for some interface methods.
"""
class multiplyForm(form):
    """
    initialize the form
    self.elements is a set that contains all the elements (individuals or forms) 
    that belong to this form
    """
    def __init__(self, sort):
        super().__init__(sort)
        self.elements = []
        self.maximal = True
        
    def __iter__(self):
        return iter(self.elements)
        
    """
    check if the form is empty
    """
    def isEmpty(self):
        return len(self.elements) == 0


    def __len__(self):
        return len(self.elements)
        
    """
    return the first element of the form
    """
    def first(self):
        if len(self.elements)==0:
            return None
        else: return self.elements[0]
        
    """
    return the last element of the form
    """
    def last(self):
        if len(self.elements)==0:
            return None
        else: return self.elements[-1]
    
    """
    return the next element in the form after current element
    """
    def next(self, current):
        if current is None:
            return None
        for i, e in enumerate(self.elements):
            if (e == current) and (i+1 < len(self.elements)):
                return self.elements[i+1]
        return None
        
    """
    return the element that is equal/same as obj
    """
    def current(self, obj):
        for i, e in enumerate(self.elements):
            if e == obj and i+1 < len(self.elements):
                return self.elements[i]
        return None
        
    """
    return the previous element in the form before current element
    """
    def previous(self, current):
        for i, e in enumerate(self.elements):
            if e.__eq__(current):
                return self.elements[i-1]
        return None
    
    """
    check if the form contains a certain element
    """
    def contains(self, current):
        for i, e in enumerate(self.elements):
            if e.__eq__(current):
                return True
        return False

    """
    check equality between two forms
    if the elements are the same between two forms, they are equal
    """
    def __eq__(self, other):
        if other is None or self.type != other.type:
            return False
        for i in self.elements:
            check = False
            for j in other.elements:
                if i == j:
                    check = True
            if check == False:
                return False
        return True
        
    """
    merging another form into this form
    i.e. inserting all elements belonging to other form into this form
    """
    def merge(self, other):
        for i in other.elements:
            self.insert(i)
            
    """
    add an element into this form at a certain index
    """
    def add(self, index, element):
        if element.assoc is None:
            self.elements.insert(index, element)
            element.assoc = self
        else:
            newCopy = deepcopy(element) #element.duplicate()
            self.elements.insert(index, newCopy)
            newCopy.assoc = self
        self.maximal = False
        
    def duplicateForm(self):
        duplicate = self.sort.newForm()
        for i in self.elements:
            duplicate.insert(i)
        return duplicate
        
    """
    remove an element from this form
    """
    def remove(self, element):
        for i in self.elements:
            if i is element or i == element:
                self.elements.remove(i)
        element.assoc = None
        return   
        
    """
    pop an element with index out of the list and return it
    index = 0 to pop the 1st element ... -1 for last element
    """
    def pop(self, index):
        return self.elements.pop(index)
        
    """
    going through this form for rule analysis or rule detection
    """
    def accept(self, visitor, direction=-1):
        if visitor.visitEnter(self) and not self.isEmpty():
            self.maximalize()
            for element in self.elements:
                element.accept(visitor, direction)
        visitor.visitLeave(self)
        return
        
    """
    editing the form in a certain procedure(editor)
    mainly use for transforming the form elements with transformation matrix(es)
    """        
    def edit(self, editor):
        self.maximal = False
        result = self.sort.newForm()
        for i in self.elements:
            result.insert(i.edit(editor))
        return result
        
    """
    converting all individuals inside this form that uses vectors with coordinates into 
    using vectors with float values 
    """
    def convertFloat(self):
        result = self.sort.newForm()
        for i in self.elements:
            result.insert(i.convertFloat())
        result.maximal = self.maximal
        return result
        
    """
    search within this form an element that when converted to float value, has the same value
    as the element to be found
    """
    def findFloatValue(self, element):
        for i in self.elements:
            if i.sort == element.sort:
                result = i.findFloatValue(element)
                if result is not False:
                    return result
        return False
        
    """
    check if this form contains a certain individual
    """
    def containsInd(self, ind):
        if self.type == 'form-metaForm':
            for i in self.elements:
                if i.sort == ind.sort or i.type == 'form-metaForm':
                    if i.containsInd(ind):
                        return True
        elif self.sort == ind.sort:
            for i in self.elements:
                if i.containsInd(ind):
                    return True
        return False
        
    """
    check if this form is contained by the compareForm after transforming
    with transformSet
    """
    def checkTransform(self, transformSet, compareForm):
        for i in self.elements:
            if not i.checkTransform(transformSet, compareForm):
                return False
        return True      
    
    """
    clear the content of the form
    """
    def clear(self):
        for i in self.elements:
            i.assoc = None
        self.elements.clear()
        
    #------------------------------------------------------------------
    
"""    
    def hashcode(self):
        return
    def toString(self):
        return
    def parse(self):
        return 
"""