# -*- coding: utf-8 -*-
"""
Created on Thu Aug 18 13:26:56 2016

@author: akibdpt
"""

from .form import form
from copy import deepcopy

"""
 * A singly-form denotes a form of a singly-associated sort.
 * Singly-forms may contain only one individual.
 * Different types of singly-forms are distinguished based on the operational
 * behavior of their forms. 
 *
 * An {OrdinalForm} specifies an ordinal behavior on numbers: any set of 
 * numbers is reduced to the single maximum number, all other numbers in the 
 * original set are a part of this maximum number.
 *
 * An singly-form is maximal if it is not equal to nil and its
 * individual's attribute form is maximal. If it is equal to nil, it
 * should not have an individual, i.e., its individual should be null.
 * 
 * The SinglyForm class extends on the {Form} class.
 * It implements the form container in a trivial way, containing a single
 * individual.
 * It provides common implementations for some interface methods.
"""
class singlyForm(form):
    """
    initialize the form
    self.individual is the single individual that belongs to this form
    """
    def __init__(self, sort):
        super().__init__(sort)
        self.individual = None
        
    def __iter__(self):
        return iter([self.individual])
        
    """
    check if the form is empty
    """
    def isEmpty(self):
        if self.individual == None:
            return True
        return self.individual.isNil()
    """
    return the first element of the form which is the only individual
    """
    def first(self):
        return self.individual

    """
    return the last element of the form which is also the only individual
    """
    def last(self):
        return self.individual

    def next(self, current):
        return None

    def previous(self, current):
        return None

    """
    maximalize the form
    if the individual is Nil(empty, or has value of 0/min): remove the individual
    if the individual has attribute, maximalize the attribute form
    """
    def maximalize(self):
        if self.individual is None:
            return
        if self.individual.isNil():
            self.individual = None
        elif self.individual.hasAttribute():
            self.individual.attribute.maximalize()
    """
    check if two forms are equal
    if the individual in the two forms are equal, the two forms are equal 
    """
    def __eq__(self, other):
        if other is None or self.type != other.type:
            return False
        return (self.individual == other.individual)
        
    """
    merge the other form into this form
    i.e. changing this form's individual into the other's individual if the 
         other's individua is greater than this form's individual
    """
    def merge(self, other):
        if other.individual.compareValue(self.individual) == 1:
            self.add(other.individual)
        return
        
    """
    make the individual into this form's individual
    """        
    def add(self, individual):
        if individual.assoc is None:
            self.individual = individual
            individual.assoc = self
        else:
            newCopy = individual.duplicateInd() #deepcopy(individual) #
            self.individual = newCopy
            newCopy.assoc = self
        
    """
    remove the individual from this form (make it minimum)
    """
    def remove(self):
        self.individual = self.individual.getMinInd()          
    
    """
    check this form's individual for rule analysis or rule detection
    """
    def accept(self, visitor, direction=-1):
        if visitor.visitEnter(self) and not self.isEmpty():
            self.maximalize()
            self.individual.accept(visitor, direction)
        visitor.visitLeave(self)
        return
        
    """
    editing the form in a certain procedure(editor)
    mainly use for transforming the form elements with transformation matrix(es)
    """        
    def edit(self, editor):
        result = self.sort.newForm()
        result.individual = self.individual.edit(editor)
        return result
        
    """
    converting all individuals inside this form that uses vectors with coordinates into 
    using vectors with float values 
    """
    def convertFloat(self):
        result = self.sort.newForm()
        result.individual = self.individual.convertFloat()
        return result
        
    """
    check when this form's individual is converted to float value, if it has 
    the same value as the element to be found
    """
    def findFloatValue(self, element):
        if self.sort == element.sort:
            return self.individual.findFloatValue(element)
        return False
        
    """
    check if this form contains a certain individual
    i.e. this form's individual is ind
    """
    def containsInd(self, ind):
        if self.individual == ind:
            return True
        return False
        
    """
    check if this form is contained by the compareForm after transforming
    with transformSet
    """
    def checkTransform(self, transformSet, compareForm):
        if self.individual.checkTransform(transformSet, compareForm):
            return True
        return False           
        
    """
    clear the content of the form
    """
    def clear(self):
        self.individual.assoc = None
        self.individual = None
    #------------------------------------------------------------------
    
"""    
    def hashcode(self):
        return
    def toString(self):
        return
    def parse(self):
        return 
    def clear(self):
        return 
"""

"""
    def insert(self):
        return
    def merge(self):
        return 
""" 
        