# -*- coding: utf-8 -*-
"""
Created on Tue Aug 23 12:22:54 2016

@author: akibdpt
"""
from .individual import individual
from copy import deepcopy
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..struct.vector import vector2D, vector3D
from ..struct.vectorFloat import vector2Dfloat, vector3Dfloat
"""
 * A line is a linear, connected, non-bounded planar curve with a
 * discrete behavior.
 * 
 * The Line class extends on the Individual class.
 * It defines the characteristic individual for lines. A line is represented as
 * a direction vector and a position vector specifying the root of the line.
 * This characteristic individual accepts no parameters.
 * It specifies a {sortal.map.similarityMap} as default.
 * Forms of lines adhere to a discrete behavior.
"""
class line2D(individual):

    """
    initialize a line
    2 points (start and end) are passed into the function.
    the direction of the line will be calculated by having end subtracting start
                                (the direction's x value must always be >= 0)
    the unit direction is also calculated by dividing direction by its length
    root is the root vector of the line
    """
    def __init__(self, sort, start, end):
        super().__init__(sort)     
        self.type = 'ind-line2D'

        if end.type == 'ind-point2D':
            end = end.value
        if start.type == 'ind-point2D':
            start = start.value
        if end.compareValue(start) == -1:
            temp = start
            start = end
            end = temp
        self.direction = (end.subtract(start))
        self.unitDir = self.direction.unit()
        self.root = start.subtract(self.direction.scale(start))
        
    def __str__(self):
        return self.type+'_'+'r:'+str(self.root)+', dir:'+str(self.direction)
                
        
    def printVal(self):
        return ('direction: '+str(self.direction)+ ', unitDir: '+
                str(self.unitDir) + ', root: '+str(self.root))
        
    def isNil(self):
        return self.direction.x == 0 and self.direction.y == 0
        
    def __eq__(self, other):
        if other is None or other.type != 'ind-line2D' :
            return False
        if self.sort == other.sort:
            return (self.direction == other.direction) and (self.root == other.root)
        
    def compareValue(self, other):
        c = self.unitDir.compareValue(other.unitDir) 
        if c != 0:
            return c
        return self.root.compareValue(other.root) # c = self.root.compareValue(other.root)

    """
    check if two lines are parallel
    """
    def parallel(self, other):
        if other is None or other.type != 'ind-line2D' :
            return False
        return self.direction.parallel(other.direction)
        
    """
    check if two lines intersect
    """
    def intersect(self, other):
        if other is None or other.type != 'ind-line2D' :
            return False
        return not self.direction.parallel(other.direction)
     
    """
    find the intersection of two lines
    """
    def intersection(self, other):
        if not self.intersect(other):
            raise ValueError('the two lines are not intersecting')
        if self.direction.x == 0:
            a2 = other.direction.y / other.direction.x
            x = self.root.x
            b2 = other.root.y - a2*other.root.x
            y = a2*x+b2
        elif other.direction.x == 0:
            a1 = self.direction.y / self.direction.x
            x = other.root.x
            b1 = self.root.y - a1*self.root.x
            y = a1*x+b1
        else:
            a1 = self.direction.y / self.direction.x
            a2 = other.direction.y / other.direction.x
            b1 = self.root.y - a1*self.root.x
            b2 = other.root.y - a2*other.root.x
            x = (b2-b1)/(a1-a2)
            y = a1*x+b1
        if self.direction.type == 'struct-vector2Dfloat':
            return vector2Dfloat(x,y)
        else:
            return vector2D(x,y)
        
    """
    check if two lines are perpendicular
    """
    def isPerpendicular(self,other):
        if other is None or other.type != 'ind-line2D' :
            return False
        return self.direction.isPerpendicular(other.direction)
        
    def duplicateInd(self):
        duplicate = line2D(self.sort, self.root.add(self.direction), self.root)
        if self.attribute is not None:
            duplicate.setAttribute(deepcopy(self.attribute))
        return duplicate    
        
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is None:
            return self.duplicateInd()
        start = self.root.transform(affine);
        end = self.root.add(self.direction).transform(affine);            
        return line2D(self.sort, start, end)
        
    def convertFloat(self):
        start = self.root.add(self.direction).convertFloat()
        end = self.root.convertFloat()
        a = line2D(self.sort, start, end)
        if self.hasAttribute():
            a.addAttribute(self.attribute.convertFloat())
        return a
        
    def findFloatValue(self, element):
        if element.type != "ind-line2D":
            return False
        if (self.unitDir.convertFloat() == element.unitDir and
            self.root.convertFloat() == element.root):
            return self
        if self.hasAttribute():
            return self.attribute.findFloatValue(element)
        return False

    def checkTransform(self, transform, compareForm):
        affine = transform.retrieveTransform(affineTransform2D)
        if affine is None:
            tfLine = self
        else:
            start = self.root.transform(affine);
            end = self.root.add(self.direction).transform(affine)         
            tfLine = line2D(self.sort, start, end)
            if self.hasAttribute():
                tfLine.addAttribute(self.attribute)
        if compareForm.containsInd(tfLine):
            return True
        return False

class line3D(individual):

    """
    initialize a line
    2 points (start and end) are passed into the function.
    the direction of the line will be calculated by having end subtracting start
                                (the direction's x value must always be >= 0)
    the unit direction is also calculated by dividing direction by its length
    root is the root vector of the line
    """
    def __init__(self, sort, start, end):
        super().__init__(sort)     
        self.type = 'ind-line3D'

        if end.type == 'ind-point3D':
            end = end.value
        if start.type == 'ind-point3D':
            start = start.value
        if end.compareValue(start) == -1:
            temp = start
            start = end
            end = temp
        self.direction = end.subtract(start)
        self.unitDir = self.direction.unit()
        self.root = start.subtract(self.direction.scale(start))
        
    def __str__(self):
        return self.type+'_'+'r:'+str(self.root)+', dir:'+str(self.direction)
                
        
    def printVal(self):
        return ('direction: '+str(self.direction)+ ', unitDir: '+
                str(self.unitDir) + ', root: '+str(self.root))
        
    def isNil(self):
        return self.direction.x == 0 and self.direction.y == 0
        
    def __eq__(self, other):
        if other is None or other.type != 'ind-line3D' :
            return False
        if self.sort == other.sort:
            return (self.direction == other.direction) and (self.root == other.root)
        
    def compareValue(self, other):
        c = self.unitDir.compareValue(other.unitDir) 
        if c != 0:
            return c
        return self.root.compareValue(other.root) # c = self.root.compareValue(other.root)

    """
    check if two lines are parallel
    """
    def parallel(self, other):
        if other is None or other.type != 'ind-line3D' :
            return False
        return self.direction.parallel(other.direction)
        
    """
    check if two lines intersect
    """
    def coPlanar(self, other):
        if other is None or other.type != 'ind-line3D' :
            return False
        return self.direction.product(other.direction).dotproduct(self.root.subtract(other.root)) == 0
        
    """
    check if two lines intersect
    """
    def intersect(self, other):
        if other is None or other.type != 'ind-line3D' :
            return False
        return self.coPlanar(other) and not self.parallel(other)
        
        
    """
    check if two lines intersect
    """
    def perpendicular(self, other):
        if other is None or other.type != 'ind-line3D' :
            return False
        return self.direction.dotproduct(other.direction) == 0
     
    """
    find the intersection of two lines
    """
    def intersection(self, other):
        if not self.intersect(other):
            raise ValueError('the two lines are not intersecting')
            
        denom = self.unitDir.product(other.unitDir)
        diff = other.root.subtract(self.root)
        numer1 = diff.product(other.unitDir)
        
        if denom.x != 0:
            x = numer1.x/denom.x
        elif denom.y != 0:
            x = numer1.y/denom.y
        else:
            x = numer1.z/denom.z
        return self.root.add(self.unitDir.scale(x))
                
    def duplicateInd(self):
        duplicate = line3D(self.sort, self.root.add(self.direction), self.root)
        if self.attribute is not None:
            duplicate.setAttribute(deepcopy(self.attribute))
        return duplicate    
        
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            return self.duplicateInd()
        start = self.root.transform(affine);
        end = self.root.add(self.direction).transform(affine);            
        return line3D(self.sort, start, end)
        
    def convertFloat(self):
        start = self.root.add(self.direction).convertFloat()
        end = self.root.convertFloat()
        a = line3D(self.sort, start, end)
        if self.hasAttribute():
            a.addAttribute(self.attribute.convertFloat())
        return a
        
    def findFloatValue(self, element):
        if element.type != "ind-line3D":
            return False
        if (self.unitDir.convertFloat() == element.unitDir and
            self.root.convertFloat() == element.root):
            return self
        if self.hasAttribute():
            return self.attribute.findFloatValue(element)
        return False

    def checkTransform(self, transform, compareForm):
        affine = transform.retrieveTransform(affineTransform3D)
        if affine is None:
            tfLine = self
        else:
            start = self.root.transform(affine);
            end = self.root.add(self.direction).transform(affine)         
            tfLine = line3D(self.sort, start, end)
            if self.hasAttribute():
                tfLine.addAttribute(self.attribute)
        if compareForm.containsInd(tfLine):
            return True
        return False
